#include "ControlWidget.h"

#include <RobotAPI/libraries/armem/server/ltm/disk/MemoryManager.h>
#include <RobotAPI/libraries/armem/server/ltm/disk/operations.h>
#include <RobotAPI/libraries/armem/server/query_proc/ltm.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <QHBoxLayout>
#include <QSpacerItem>
#include <QFileDialog>
#include <QPushButton>

#include <filesystem>

#include <cmath>


namespace armarx::armem::gui::disk
{

    ControlWidget::ControlWidget()
    {
        _latestDirectory = QString::fromStdString("/tmp/MemoryExport");

        _loadFromDiskButton = new QPushButton(" Load from Disk", this);
        _loadFromDiskButton->setIcon(QIcon(":/icons/document-open.svg"));
        _storeOnDiskButton = new QPushButton(" Store on Disk", this);
        _storeOnDiskButton->setIcon(QIcon(":/icons/document-save.svg"));

        // Allow horizontal shrinking of buttons
        std::vector<QPushButton*> buttons { _storeOnDiskButton, _loadFromDiskButton };
        for (QPushButton* button : buttons)
        {
            button->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Fixed);
        }

        this->setSizePolicy(QSizePolicy::Policy::Preferred, QSizePolicy::Policy::Fixed);

        QHBoxLayout* layout = new QHBoxLayout();
        this->setLayout(layout);

        const int margin = 0;
        layout->setContentsMargins(margin, margin, margin, margin);

        layout->addWidget(_loadFromDiskButton);
        layout->addWidget(_storeOnDiskButton);


        // Connections

        connect(_loadFromDiskButton, &QPushButton::pressed, [this]()
        {
            QString directory = chooseDirectoryDialog();
            if (directory.size() > 0)
            {
                emit requestedLoadFromDisk(directory);
            }
        });
        connect(_storeOnDiskButton, &QPushButton::pressed, [this]()
        {
            QString directory = chooseDirectoryDialog();
            if (directory.size() > 0)
            {
                emit requestedStoreOnDisk(directory);
            }
        });
    }


    static
    const std::string&
    handleSingular(int num, const std::string& singular, const std::string& plural)
    {
        return num == 1 ? singular : plural;
    }


    void
    ControlWidget::storeOnDisk(
        QString directory,
        const std::map<std::string, wm::Memory> memoryData,
        std::string* outStatus)
    {
        std::filesystem::path path(directory.toUtf8().constData());
        ARMARX_CHECK_POSITIVE(path.string().size());  // An empty path indicates an error.

        std::stringstream status;
        if (std::filesystem::is_regular_file(path))
        {
            status << "Could not export memories contents to " << path << ": Cannot overwrite existing file.";
        }
        else
        {
            int numStored = 0;
            for (const auto& [name, data] : memoryData)
            {
                if (std::filesystem::is_regular_file(path / name))
                {
                    status << "Could not export memory '" << name << "' to " << path << ": Cannot overwrite existing file.\n";
                }
                else
                {
                    std::filesystem::create_directories(path / name);

                    armem::server::ltm::disk::MemoryManager manager;
                    manager.setName(name);
                    manager.setBasePath(path / name);
                    manager.reload();
                    manager.append(data);

                    numStored++;
                }
            }
            status << "Exported " << numStored << " " << handleSingular(numStored, "memory", "memories") << " to " << path << ".";
        }

        if (outStatus)
        {
            *outStatus = status.str();
        }
    }


    std::map<std::string, wm::Memory>
    ControlWidget::loadFromDisk(
        QString directory,
        const armem::client::QueryInput& _queryInput,
        std::string* outStatus)
    {
        std::filesystem::path path(directory.toUtf8().constData());

        std::map<std::string, wm::Memory> memoryData;
        std::stringstream status;

        if (not std::filesystem::is_directory(path))
        {
            status << "Could not import a memory from " << path << ". Skipping import.";
        }
        else
        {
            // We use LTM as query target for the disk
            armem::client::QueryInput queryInput = _queryInput;
            queryInput.addQueryTargetToAll(armem::query::data::QueryTarget::LTM);
            const query::data::Input queryIce = queryInput.toIce();

            int numLoaded = 0;
            for (const auto& dir : std::filesystem::directory_iterator(path))
            {
                if (dir.is_directory())
                {
                    std::string memoryName = dir.path().filename();
                    armem::wm::Memory memory = armem::server::ltm::disk::load(dir.path());
                    memoryData[memoryName] = std::move(memory);

                    numLoaded++;
                }
            }
            status << "Loaded " << numLoaded << " " << handleSingular(numLoaded, "memory", "memories") << " from " << path << ".";
        }

        if (outStatus)
        {
            *outStatus = status.str();
        }
        return memoryData;
    }


    QString ControlWidget::chooseDirectoryDialog()
    {
        QFileDialog dialog;
        dialog.setFileMode(QFileDialog::DirectoryOnly);
        dialog.setOption(QFileDialog::ShowDirsOnly, false);
        dialog.setDirectory(_latestDirectory);
        if (dialog.exec())
        {
            _latestDirectory = dialog.directory().path();
            return _latestDirectory;
        }
        else
        {
            return QString::fromStdString("");
        }
    }

}

