/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>
#include <numeric>

// Eigen
#include <Eigen/Geometry>
#include <Eigen/Core>
#include <Eigen/src/Core/util/Constants.h>

// ArmarX
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/core/navigator/data/complex/NDArray.h>


namespace armarx::aron::converter
{
    class AronEigenConverter
    {
        AronEigenConverter() = delete;

    public:
        // TODO: Add inplace methods

        static Eigen::Quaternion<float> ConvertToQuaternionf(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Quaternion<float> ConvertToQuaternionf(const datanavigator::NDArrayNavigator&);
        static Eigen::Quaternion<double> ConvertToQuaterniond(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Quaternion<double> ConvertToQuaterniond(const datanavigator::NDArrayNavigator&);
        static Eigen::Vector3f ConvertToVector3f(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Vector3f ConvertToVector3f(const datanavigator::NDArrayNavigator&);
        static Eigen::Vector3d ConvertToVector3d(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Vector3d ConvertToVector3d(const datanavigator::NDArrayNavigator&);
        static Eigen::Matrix4f ConvertToMatrix4f(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Matrix4f ConvertToMatrix4f(const datanavigator::NDArrayNavigator&);
        static Eigen::Matrix4d ConvertToMatrix4d(const datanavigator::NDArrayNavigatorPtr&);
        static Eigen::Matrix4d ConvertToMatrix4d(const datanavigator::NDArrayNavigator&);


        template<typename T>
        static Eigen::Quaternion<T> ConvertToQuaternion(const datanavigator::NDArrayNavigatorPtr& nav)
        {
            ARMARX_CHECK_NOT_NULL(nav);
            return ConvertToQuaternion<T>(*nav);
        }

        template<typename T>
        static Eigen::Quaternion<T> ConvertToQuaternion(const datanavigator::NDArrayNavigator& nav)
        {
            checkDimensions(nav, {1, 4, sizeof(T)}, "ConvertToQuaternion");
            auto dims = nav.getDimensions();

            Eigen::Quaternion<T> ret;
            memcpy(reinterpret_cast<unsigned char*>(ret.coeffs().data()), nav.getData(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()));
            return ret;
        }


        template<typename T, int Size>
        static Eigen::Matrix<T, Size, 1> ConvertToVector(const datanavigator::NDArrayNavigatorPtr& nav)
        {
            ARMARX_CHECK_NOT_NULL(nav);
            return ConvertToVector<T, Size>(*nav);
        }

        template<typename T, int Size>
        static Eigen::Matrix<T, Size, 1> ConvertToVector(const datanavigator::NDArrayNavigator& nav)
        {
            checkDimensions(nav, {Size, 1, sizeof(T)}, "ConvertToVector");
            auto dims = nav.getDimensions();

            Eigen::Matrix<T, Size, 1> ret;
            memcpy(reinterpret_cast<unsigned char*>(ret.data()), nav.getData(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()));
            return ret;
        }


        template<typename T, int Rows, int Cols>
        static Eigen::Matrix<T, Rows, Cols> ConvertToMatrix(const datanavigator::NDArrayNavigatorPtr& nav)
        {
            ARMARX_CHECK_NOT_NULL(nav);
            return ConvertToMatrix<T, Rows, Cols>(*nav);
        }

        template<typename T>
        static Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic> ConvertToDynamicMatrix(const datanavigator::NDArrayNavigator& nav)
        {
            const auto dims = nav.getDimensions();

            using MatrixT = Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic>;

            MatrixT ret;
            memcpy(reinterpret_cast<unsigned char*>(ret.data()), nav.getData(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<>()));
            return ret;
        }

        template<typename T, int Rows = Eigen::Dynamic, int Cols = Eigen::Dynamic>
        static Eigen::Matrix<T, Rows, Cols> ConvertToMatrix(const datanavigator::NDArrayNavigator& nav)
        {
            if constexpr(Rows == Eigen::Dynamic and Cols == Eigen::Dynamic)
            {
                return ConvertToDynamicMatrix<T>(nav);
            }

            checkDimensions(nav, {Rows, Cols, sizeof(T)}, "ConvertToMatrix");
            auto dims = nav.getDimensions();

            Eigen::Matrix<T, Rows, Cols> ret;
            memcpy(reinterpret_cast<unsigned char*>(ret.data()), nav.getData(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()));
            return ret;
        }

        template<typename T>
        static datanavigator::NDArrayNavigatorPtr ConvertFromMatrix(const Eigen::Matrix < T, Eigen::Dynamic, Eigen::Dynamic >& mat)
        {
            datanavigator::NDArrayNavigatorPtr ndArr(new datanavigator::NDArrayNavigator);

            ndArr->setDimensions({static_cast<int>(mat.rows()), static_cast<int>(mat.cols())});
            ndArr->setData(sizeof(T) * mat.size(), reinterpret_cast <const unsigned char* >(mat.data()));

            return ndArr;
        }


        // Eigen::Array

        template<typename T>
        static Eigen::Array<T, Eigen::Dynamic, Eigen::Dynamic> ConvertToDynamicArray(const datanavigator::NDArrayNavigator& nav)
        {
            const auto dims = nav.getDimensions();
            ARMARX_CHECK_EQUAL(dims.size(), 2);

            using ArrayT = Eigen::Array<T, Eigen::Dynamic, Eigen::Dynamic>;

            ArrayT ret = Eigen::Map<ArrayT>(reinterpret_cast<T*>(nav.getData()), dims.at(0), dims.at(1));
            return ret;
        }

        template<typename T, int Rows = Eigen::Dynamic, int Cols = Eigen::Dynamic>
        static Eigen::Matrix<T, Rows, Cols> ConvertToArray(const datanavigator::NDArrayNavigator& nav)
        {
            if constexpr(Rows == Eigen::Dynamic and Cols == Eigen::Dynamic)
            {
                return ConvertToDynamicArray<T>(nav);
            }

            checkDimensions(nav, {Rows, Cols, sizeof(T)}, "ConvertToMatrix");
            auto dims = nav.getDimensions();

            Eigen::Array<T, Rows, Cols> ret;
            memcpy(reinterpret_cast<unsigned char*>(ret.data()), nav.getData(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()));
            return ret;
        }

        template<typename T>
        static datanavigator::NDArrayNavigatorPtr ConvertFromArray(const Eigen::Array < T, Eigen::Dynamic, Eigen::Dynamic >& mat)
        {
            datanavigator::NDArrayNavigatorPtr ndArr(new datanavigator::NDArrayNavigator);

            ndArr->setDimensions({static_cast<int>(mat.rows()), static_cast<int>(mat.cols())});
            ndArr->setData(sizeof(T) * mat.size(), reinterpret_cast <const unsigned char* >(mat.data()));

            return ndArr;
        }



    private:

        /**
         * @throw `error::AronException` If `nav`'s dimensions do not match `expected`.
         */
        static void checkDimensions(const datanavigator::NDArrayNavigator& nav, const std::vector<int>& expected,
                                    const std::string& method, const std::string& caller = "AronEigenConverter");

    };

}
